# coding: utf-8
from enum import Enum
import json
from supervisely.task.progress import Progress


class PullPolicy(Enum):
    def __str__(self):
        return str(self.value)

    ALWAYS = 'Always'.lower()
    IF_AVAILABLE = 'IfAvailable'.lower()
    IF_NOT_PRESENT = 'IfNotPresent'.lower()
    NEVER = 'Never'.lower()


def docker_pull_if_needed(docker_api, docker_image_name, policy, logger, progress=True):
    logger.info('docker_pull_if_needed args', extra={
        'policy': policy, 
        'type(policy)': type(policy),
        'policy == PullPolicy.ALWAYS': str(policy) == str(PullPolicy.ALWAYS),
        'policy == PullPolicy.NEVER': str(policy) == str(PullPolicy.NEVER),
        'policy == PullPolicy.IF_NOT_PRESENT': str(policy) == str(PullPolicy.IF_NOT_PRESENT),
        'policy == PullPolicy.IF_AVAILABLE': str(policy) == str(PullPolicy.IF_AVAILABLE)
    })
    if str(policy) == str(PullPolicy.ALWAYS):
        if progress is False:
            _docker_pull(docker_api, docker_image_name, logger)
        else:
            _docker_pull_progress(docker_api, docker_image_name, logger)
    elif str(policy) == str(PullPolicy.NEVER):
        pass
    elif str(policy) == str(PullPolicy.IF_NOT_PRESENT):
        if not _docker_image_exists(docker_api, docker_image_name):
            if progress is False:
                _docker_pull(docker_api, docker_image_name, logger)
            else:
                _docker_pull_progress(docker_api, docker_image_name, logger)
    elif str(policy) == str(PullPolicy.IF_AVAILABLE):
        if progress is False:
            _docker_pull(docker_api, docker_image_name, logger, raise_exception=False)
        else:
            _docker_pull_progress(docker_api, docker_image_name, logger, raise_exception=False)
    else:
        raise RuntimeError(f"Unknown pull policy {str(policy)}")
    if not _docker_image_exists(docker_api, docker_image_name):
        raise RuntimeError(f"Docker image {docker_image_name} not found. Agent's PULL_POLICY is {str(policy)}")


def _docker_pull(docker_api, docker_image_name, logger, raise_exception=True):
    from docker.errors import DockerException
    logger.info('Docker image will be pulled', extra={'image_name': docker_image_name})
    progress_dummy = Progress('Pulling image...', 1, ext_logger=logger)
    progress_dummy.iter_done_report()
    try:
        pulled_img = docker_api.images.pull(docker_image_name)
        logger.info('Docker image has been pulled', extra={'pulled': {'tags': pulled_img.tags, 'id': pulled_img.id}})
    except DockerException as e:
        if raise_exception is True:
            raise DockerException('Unable to pull image: see actual error above. '
                                  'Please, run the task again or contact support team.')
        else:
            logger.warn("Pulling step is skipped. Unable to pull image: {!r}.".format(str(e)))


def _docker_pull_progress(docker_api, docker_image_name, logger, raise_exception=True):
    logger.info('Docker image will be pulled', extra={'image_name': docker_image_name})
    from docker.errors import DockerException
    try:
        layers_total = {}
        layers_current = {}
        progress = Progress('Pulling dockerimage', 1, is_size=True, ext_logger=logger)
        for line in docker_api.api.pull(docker_image_name, stream=True, decode=True):
            layer_id = line.get("id", None)
            progress_details = line.get("progressDetail", {})
            if "total" in progress_details and "current" in progress_details:
                layers_total[layer_id] = progress_details["total"]
                layers_current[layer_id] = progress_details["current"]
                total = sum(layers_total.values())
                current = sum(layers_current.values())
                if total > progress.total:
                    progress.set(current, total)
                    progress.report_progress()
                elif (current - progress.current) / total > 0.01:
                    progress.set(current, total)
                    progress.report_progress()

            #print(json.dumps(line, indent=4))
        logger.info('Docker image has been pulled', extra={'image_name': docker_image_name})
    except DockerException as e:
        if raise_exception is True:
            raise DockerException('Unable to pull image: see actual error above. '
                                  'Please, run the task again or contact support team.')
        else:
            logger.warn("Pulling step is skipped. Unable to pull image: {!r}.".format(repr(e)))


def _docker_image_exists(docker_api, docker_image_name):
    from docker.errors import ImageNotFound
    try:
        docker_img = docker_api.images.get(docker_image_name)
    except ImageNotFound:
        return False
    return True


# coding: utf-8
"""download/upload/manipulate neural networks"""

import os
import tarfile
from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor
import numpy as np
import json

from supervisely.api.module_api import ApiField, CloneableModuleApi, RemoveableModuleApi
from supervisely._utils import rand_str
from supervisely.io.fs import ensure_base_path, silent_remove
from supervisely.imaging import image as sly_image
from supervisely.project.project_meta import ProjectMeta


class NeuralNetworkApi(CloneableModuleApi, RemoveableModuleApi):
    @staticmethod
    def info_sequence():
        return [ApiField.ID,
                ApiField.NAME,
                ApiField.DESCRIPTION,
                ApiField.CONFIG,
                ApiField.HASH,
                ApiField.ONLY_TRAIN,
                ApiField.PLUGIN_ID,
                ApiField.PLUGIN_VERSION,
                ApiField.SIZE,
                ApiField.WEIGHTS_LOCATION,
                ApiField.README,
                ApiField.TASK_ID,
                ApiField.USER_ID,
                ApiField.TEAM_ID,
                ApiField.WORKSPACE_ID,
                ApiField.CREATED_AT,
                ApiField.UPDATED_AT]

    @staticmethod
    def info_tuple_name():
        return 'ModelInfo'

    def get_list(self, workspace_id, filters=None):
        return self.get_list_all_pages('models.list',  {ApiField.WORKSPACE_ID: workspace_id, ApiField.FILTER: filters or []})

    def get_info_by_id(self, id):
        return self._get_info_by_id(id, 'models.info')

    def download(self, id):
        response = self._api.post('models.download', {ApiField.ID: id}, stream=True)
        return response

    def download_to_tar(self, workspace_id, name, tar_path, progress_cb=None):
        model = self.get_info_by_name(workspace_id, name)
        response = self.download(model.id)
        ensure_base_path(tar_path)
        with open(tar_path, 'wb') as fd:
            for chunk in response.iter_content(chunk_size=1024*1024):
                fd.write(chunk)
                if progress_cb is not None:
                    read_mb = len(chunk) / 1024.0 / 1024.0
                    progress_cb(read_mb)

    def download_to_dir(self, workspace_id, name, directory, progress_cb=None):
        model_tar = os.path.join(directory, rand_str(10) + '.tar')
        self.download_to_tar(workspace_id, name, model_tar, progress_cb)
        model_dir = os.path.join(directory, name)
        with tarfile.open(model_tar) as archive:
            archive.extractall(model_dir)
        silent_remove(model_tar)
        return model_dir

    def generate_hash(self, task_id):
        response = self._api.post('models.hash.create', {ApiField.TASK_ID: task_id})
        return response.json()

    def upload(self, hash, archive_path, progress_cb=None):
        encoder = MultipartEncoder({'hash': hash,
                                    'weights': (os.path.basename(archive_path), open(archive_path, 'rb'), 'application/x-tar')})

        def callback(monitor_instance):
            read_mb = monitor_instance.bytes_read / 1024.0 / 1024.0
            if progress_cb is not None:
                progress_cb(read_mb)
        monitor = MultipartEncoderMonitor(encoder, callback)
        self._api.post('models.upload', monitor)

    def inference_remote_image(self, id, image_hash, ann=None, meta=None, mode=None):
        data = {
            "request_type": "inference",
            "meta": meta or ProjectMeta().to_json(),
            "annotation": ann or None,
            "mode": mode or {},
            "image_hash": image_hash
        }
        fake_img_data = sly_image.write_bytes(np.zeros([5, 5, 3]), '.jpg')
        encoder = MultipartEncoder({'id': str(id).encode('utf-8'),
                                    'data': json.dumps(data),
                                    'image': ("img", fake_img_data, "")})
        response = self._api.post('models.infer', MultipartEncoderMonitor(encoder))
        return response.json()

    def inference(self, id, img, ann=None, meta=None, mode=None, ext=None):
        data = {
            "request_type": "inference",
            "meta": meta or ProjectMeta().to_json(),
            "annotation": ann or None,
            "mode": mode or {},
        }
        img_data = sly_image.write_bytes(img, ext or '.jpg')
        encoder = MultipartEncoder({'id': str(id).encode('utf-8'),
                                    'data': json.dumps(data),
                                    'image': ("img", img_data, "")})

        response = self._api.post('models.infer', MultipartEncoderMonitor(encoder))
        return response.json()

    def get_output_meta(self, id, input_meta=None, inference_mode=None):
        data = {
            "request_type": "get_out_meta",
            "meta": input_meta or ProjectMeta().to_json(),
            "mode": inference_mode or {}
        }
        encoder = MultipartEncoder({'id': str(id).encode('utf-8'),
                                    'data': json.dumps(data)})
        response = self._api.post('models.infer', MultipartEncoderMonitor(encoder))
        response_json = response.json()
        if 'out_meta' in response_json:
            return response_json['out_meta']
        if 'output_meta' in response_json:
            return response_json['output_meta']
        return response.json()

    def get_deploy_tasks(self, model_id):
        response = self._api.post('models.info.deployed', {'id': model_id})
        return [task[ApiField.ID] for task in response.json()]

    def get_training_metrics(self, model_id):
        response = self._get_response_by_id(id=model_id, method='tasks.train-metrics', id_field=ApiField.MODEL_ID)
        return response.json() if (response is not None) else None

    def _clone_api_method_name(self):
        return 'models.clone'

    def _remove_api_method_name(self):
        return 'models.remove'

    def create_from_checkpoint(self, task_id, checkpoint_id, model_name, change_name_if_conflict=True):
        # FYI: checkpoint has these fields
        # 'modelTitle': 'my_model_name_006',
        # 'status': 'uploaded'

        self._api.task._validate_checkpoints_support(task_id)
        task_info = self._api.task.get_info_by_id(task_id)
        workspace_id = task_info[ApiField.WORKSPACE_ID]
        new_model_name = self.get_free_name(workspace_id, model_name)
        if new_model_name != model_name and change_name_if_conflict is False:
            raise KeyError("Model name={!r} already exists in workspace id={!r}".format(model_name, workspace_id))
        resp = self._api.post("models.create-from-checkpoint", {ApiField.ID: checkpoint_id,
                                                                ApiField.TASK_ID: task_id,
                                                                ApiField.NAME: new_model_name})
        process_task_id = resp.json()[ApiField.TASK_ID]
        if process_task_id is not None:
            self._api.task.wait(process_task_id, self._api.task.Status.FINISHED)
        else:
            # upload process skipped because checkpoint is already uploaded to server, just new model will be created
            pass
        return new_model_name


# coding: utf-8

# docs
from __future__ import annotations
from typing import List, Tuple, Dict, Optional
from supervisely.geometry.image_rotator import ImageRotator
from skimage import morphology as skimage_morphology

import base64
from enum import Enum
import zlib
import io
import cv2
import numpy as np
from distutils.version import StrictVersion

from PIL import Image

from supervisely.geometry.bitmap_base import BitmapBase, resize_origin_and_bitmap
from supervisely.geometry.point_location import PointLocation, row_col_list_to_points
from supervisely.geometry.polygon import Polygon
from supervisely.geometry.rectangle import Rectangle
from supervisely.geometry.constants import BITMAP


class SkeletonizeMethod(Enum):
    """
    Specifies possible skeletonization methods of :class:`Bitmap<Bitmap>`.
    """
    SKELETONIZE = 0
    MEDIAL_AXIS = 1
    THINNING = 2


def _find_mask_tight_bbox(raw_mask: np.ndarray) -> Rectangle:
    rows = list(np.any(raw_mask, axis=1).tolist())  # Redundant conversion to list to help PyCharm static analysis.
    cols = list(np.any(raw_mask, axis=0).tolist())
    top_margin = rows.index(True)
    bottom_margin = rows[::-1].index(True)
    left_margin = cols.index(True)
    right_margin = cols[::-1].index(True)
    return Rectangle(top=top_margin, left=left_margin, bottom=len(rows) - 1 - bottom_margin,
                     right=len(cols) - 1 - right_margin)


class Bitmap(BitmapBase):
    """
    Bitmap geometry for a single :class:`Label<supervisely.annotation.label.Label>`. :class:`Bitmap<Bitmap>` object is immutable.

    :param data: Bitmap mask data.
    :type data: np.ndarray
    :param origin: :class:`PointLocation<supervisely.geometry.point_location.PointLocation>`: top, left corner of Bitmap. Position of the Bitmap within image.
    :type origin: PointLocation, optional
    :param sly_id: Bitmap ID in Supervisely server.
    :type sly_id: int, optional
    :param class_id: ID of :class:`ObjClass<supervisely.annotation.obj_class.ObjClass>` to which Bitmap belongs.
    :type class_id: int, optional
    :param labeler_login: Login of the user who created Bitmap.
    :type labeler_login: str, optional
    :param updated_at: Date and Time when Bitmap was modified last. Date Format: Year:Month:Day:Hour:Minute:Seconds. Example: '2021-01-22T19:37:50.158Z'.
    :type updated_at: str, optional
    :param created_at: Date and Time when Bitmap was created. Date Format is the same as in "updated_at" parameter.
    :type created_at: str, optional
    :raises: :class:`ValueError`, if data is not bool or no pixels set to True in data
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        # Create simple bitmap
        mask = np.array([[0, 0, 0, 0, 0],
                         [0, 1, 1, 1, 0],
                         [0, 1, 0, 1, 0],
                         [0, 1, 1, 1, 0],
                         [0, 0, 0, 0, 0]], dtype=np.bool_)

        figure = sly.Bitmap(mask)

        # Note, when creating a bitmap, the specified mask is cut off by positive values, based on this, a origin is formed:
        print(figure.data)
        # Output:
        #    [[ True  True  True]
        #     [ True False  True]
        #     [ True  True  True]]

        origin = figure.origin.to_json()
        print(json.dumps(origin, indent=4))
        # Output: {
        #     "points": {
        #         "exterior": [
        #             [
        #                 1,
        #                 1
        #             ]
        #         ],
        #         "interior": []
        #     }

        # Create bitmap from black and white image:
        img = sly.imaging.image.read(os.path.join(os.getcwd(), 'black_white.jpeg'))
        mask = img[:, :, 0].astype(bool) # Get 2-dimensional bool numpy array
        figure = sly.Bitmap(mask)

     .. image:: https://i.imgur.com/2L3HRPs.jpg
    """
    @staticmethod
    def geometry_name():
        return 'bitmap'

    def __init__(self, data: np.ndarray, origin: Optional[PointLocation] = None,
                 sly_id: Optional[int] = None, class_id: Optional[int] = None, labeler_login: Optional[int] = None,
                 updated_at: Optional[str] = None, created_at: Optional[str] = None):
        if data.dtype != np.bool:
            raise ValueError('Bitmap mask data must be a boolean numpy array. Instead got {}.'.format(str(data.dtype)))

        # Call base constructor first to do the basic dimensionality checks.
        super().__init__(data=data, origin=origin, expected_data_dims=2,
                         sly_id=sly_id, class_id=class_id,
                         labeler_login=labeler_login, updated_at=updated_at, created_at=created_at)

        # Crop the empty margins of the boolean mask right away.
        if not np.any(data):
            raise ValueError('Creating a bitmap with an empty mask (no pixels set to True) is not supported.')
        data_tight_bbox = _find_mask_tight_bbox(self._data)
        self._origin = self._origin.translate(drow=data_tight_bbox.top, dcol=data_tight_bbox.left)
        self._data = data_tight_bbox.get_cropped_numpy_slice(self._data)

    @classmethod
    def _impl_json_class_name(cls):
        return BITMAP

    def rotate(self, rotator: ImageRotator) -> Bitmap:
        """
        Rotates current Bitmap.

        :param rotator: :class:`ImageRotator<supervisely.geometry.image_rotator.ImageRotator>` for Bitamp rotation.
        :type rotator: ImageRotator
        :return: Bitmap object
        :rtype: :class:`Bitmap<Bitmap>`

        :Usage Example:

         .. code-block:: python

            import supervisely as sly
            from supervisely.geometry.image_rotator import ImageRotator

            height, width = ann.img_size
            rotator = ImageRotator((height, width), 25)
            # Remember that Bitmap class object is immutable, and we need to assign new instance of Bitmap to a new variable
            rotate_figure = figure.rotate(rotator)
        """
        full_img_mask = np.zeros(rotator.src_imsize, dtype=np.uint8)
        self.draw(full_img_mask, 1)
        # TODO this may break for one-pixel masks (it can disappear during rotation). Instead, rotate every pixel
        #  individually and set it in the resulting bitmap.
        new_mask = rotator.rotate_img(full_img_mask, use_inter_nearest=True).astype(np.bool)
        return Bitmap(data=new_mask)

    def crop(self, rect: Rectangle) -> List[Bitmap]:
        """
        Crops current Bitmap.

        :param rect: Rectangle object for cropping.
        :type rect: Rectangle
        :return: List of Bitmaps
        :rtype: :class:`List[Bitmap]<supervisely.geometry.bitmap.Bitmap>`

        :Usage Example:

         .. code-block:: python

            crop_figures = figure.crop(sly.Rectangle(1, 1, 300, 350))
        """
        maybe_cropped_bbox = self.to_bbox().crop(rect)
        if len(maybe_cropped_bbox) == 0:
            return []
        else:
            [cropped_bbox] = maybe_cropped_bbox
            cropped_bbox_relative = cropped_bbox.translate(drow=-self.origin.row, dcol=-self.origin.col)
            cropped_mask = cropped_bbox_relative.get_cropped_numpy_slice(self._data)
            if not np.any(cropped_mask):
                return []
            return [Bitmap(data=cropped_mask, origin=PointLocation(row=cropped_bbox.top, col=cropped_bbox.left))]

    def resize(self, in_size: Tuple[int, int], out_size: Tuple[int, int]) -> Bitmap:
        """
        Resizes current Bitmap.

        :param in_size: Input image size (height, width) to which Bitmap belongs.
        :type in_size: Tuple[int, int]
        :param out_size: Output image size (height, width) to which Bitmap belongs.
        :type out_size: Tuple[int, int]
        :return: Bitmap object
        :rtype: :class:`Bitmap<Bitmap>`

        :Usage Example:

         .. code-block:: python

            in_height, in_width = 800, 1067
            out_height, out_width = 600, 800
            # Remember that Bitmap class object is immutable, and we need to assign new instance of Bitmap to a new variable
            resize_figure = figure.resize((in_height, in_width), (out_height, out_width))
        """
        scaled_origin, scaled_data = resize_origin_and_bitmap(self._origin, self._data.astype(np.uint8), in_size,
                                                              out_size)
        # TODO this might break if a sparse mask is resized too thinly. Instead, resize every pixel individually and set
        #  it in the resulting bitmap.
        return Bitmap(data=scaled_data.astype(np.bool), origin=scaled_origin)

    def _draw_impl(self, bitmap, color, thickness=1, config=None):
        self.to_bbox().get_cropped_numpy_slice(bitmap)[self.data] = color

    def _draw_contour_impl(self, bitmap, color, thickness=1, config=None):
        if StrictVersion(cv2.__version__) >= StrictVersion('4.0.0'):
            contours, _ = cv2.findContours(self.data.astype(np.uint8), cv2.RETR_LIST, cv2.CHAIN_APPROX_SIMPLE)
        else:
            _, contours, _ = cv2.findContours(self.data.astype(np.uint8), cv2.RETR_LIST, cv2.CHAIN_APPROX_SIMPLE)
        if contours is not None:
            for cont in contours:
                cont[:, :] += (self.origin.col, self.origin.row)  # cont with shape (rows, ?, 2)
            cv2.drawContours(bitmap, contours, -1, color, thickness=thickness)

    @property
    def area(self) -> float:
        """
        Bitmap area.

        :return: Area of current Bitmap
        :rtype: :class:`float`
        :Usage example:

         .. code-block:: python

            print(figure.area)
            # Output: 88101.0
        """
        return float(self._data.sum())

    @staticmethod
    def base64_2_data(s: str) -> np.ndarray:
        """
        Convert base64 encoded string to numpy array.

        :param s: Input base64 encoded string.
        :type s: str
        :return: Bool numpy array
        :rtype: :class:`np.ndarray`
        :Usage example:

         .. code-block:: python

              import supervisely as sly

              encoded_string = 'eJzrDPBz5+WS4mJgYOD19HAJAtLMIMwIInOeqf8BUmwBPiGuQPr///9Lb86/C2QxlgT5BTM4PLuRBuTwebo4hlTMSa44sKHhISMDuxpTYrr03F6gDIOnq5/LOqeEJgDM5ht6'
              figure_data = sly.Bitmap.base64_2_data(encoded_string)
              print(figure_data)
              #  [[ True  True  True]
              #   [ True False  True]
              #   [ True  True  True]]
        """
        z = zlib.decompress(base64.b64decode(s))
        n = np.frombuffer(z, np.uint8)

        imdecoded = cv2.imdecode(n, cv2.IMREAD_UNCHANGED)
        if (len(imdecoded.shape) == 3) and (imdecoded.shape[2] >= 4):
            mask = imdecoded[:, :, 3].astype(bool)  # 4-channel imgs
        elif len(imdecoded.shape) == 2:
            mask = imdecoded.astype(bool)  # flat 2d mask
        else:
            raise RuntimeError('Wrong internal mask format.')
        return mask

    @staticmethod
    def data_2_base64(mask: np.ndarray) -> str:
        """
        Convert numpy array to base64 encoded string.

        :param mask: Bool numpy array.
        :type mask: np.ndarray
        :return: Base64 encoded string
        :rtype: :class:`str`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            address = 'https://app.supervise.ly/'
            token = 'Your Supervisely API Token'
            api = sly.Api(address, token)

            # Get annotation from API
            meta_json = api.project.get_meta(PROJECT_ID)
            meta = sly.ProjectMeta.from_json(meta_json)
            ann_info = api.annotation.download(IMAGE_ID)
            ann = sly.Annotation.from_json(ann_info.annotation, meta)

            # Get Bitmap from annotation
            for label in ann.labels:
                if type(label.geometry) == sly.Bitmap:
                    figure = label.geometry

            encoded_string = sly.Bitmap.data_2_base64(figure.data)
            print(encoded_string)
            # 'eJzrDPBz5+WS4mJgYOD19HAJAtLMIMwIInOeqf8BUmwBPiGuQPr///9Lb86/C2QxlgT5BTM4PLuRBuTwebo4hlTMSa44sKHhISMDuxpTYrr03F6gDIOnq5/LOqeEJgDM5ht6'
        """
        img_pil = Image.fromarray(np.array(mask, dtype=np.uint8))
        img_pil.putpalette([0, 0, 0, 255, 255, 255])
        bytes_io = io.BytesIO()
        img_pil.save(bytes_io, format='PNG', transparency=0, optimize=0)
        bytes_enc = bytes_io.getvalue()
        return base64.b64encode(zlib.compress(bytes_enc)).decode('utf-8')

    def skeletonize(self, method_id: SkeletonizeMethod) -> Bitmap:
        """
        Compute the skeleton, medial axis transform or morphological thinning of Bitmap.

        :param method_id: Method to convert bool numpy array.
        :type method_id: SkeletonizeMethod
        :return: Bitmap object
        :rtype: :class:`Bitmap<Bitmap>`
        :Usage example:

         .. code-block:: python

            # Remember that Bitmap class object is immutable, and we need to assign new instance of Bitmap to a new variable
            skeleton_figure = figure.skeletonize(SkeletonizeMethod.SKELETONIZE)
            med_ax_figure = figure.skeletonize(SkeletonizeMethod.MEDIAL_AXIS)
            thin_figure = figure.skeletonize(SkeletonizeMethod.THINNING)
        """
        if method_id == SkeletonizeMethod.SKELETONIZE:
            method = skimage_morphology.skeletonize
        elif method_id == SkeletonizeMethod.MEDIAL_AXIS:
            method = skimage_morphology.medial_axis
        elif method_id == SkeletonizeMethod.THINNING:
            method = skimage_morphology.thin
        else:
            raise NotImplementedError('Method {!r} does\'t exist.'.format(method_id))

        mask_u8 = self.data.astype(np.uint8)
        res_mask = method(mask_u8).astype(bool)
        return Bitmap(data=res_mask, origin=self.origin)

    def to_contours(self) -> List[Polygon]:
        """
        Get list of contours in Bitmap.

        :return: List of Polygon objects
        :rtype: :class:`List[Polygon]<supervisely.geometry.polygon.Polygon>`
        :Usage example:

         .. code-block:: python

            figure_contours = figure.to_contours()
        """
        origin, mask = self.origin, self.data
        if StrictVersion(cv2.__version__) >= StrictVersion('4.0.0'):
            contours, hier = cv2.findContours(
                mask.astype(np.uint8),
                mode=cv2.RETR_CCOMP,  # two-level hierarchy, to get polygons with holes
                method=cv2.CHAIN_APPROX_SIMPLE
            )
        else:
            _, contours, hier = cv2.findContours(
                mask.astype(np.uint8),
                mode=cv2.RETR_CCOMP,  # two-level hierarchy, to get polygons with holes
                method=cv2.CHAIN_APPROX_SIMPLE
            )
        if (hier is None) or (contours is None):
            return []

        res = []
        for idx, hier_pos in enumerate(hier[0]):
            next_idx, prev_idx, child_idx, parent_idx = hier_pos
            if parent_idx < 0:
                external = contours[idx][:, 0].tolist()
                internals = []
                while child_idx >= 0:
                    internals.append(contours[child_idx][:, 0])
                    child_idx = hier[0][child_idx][0]
                if len(external) > 2:
                    new_poly = Polygon(exterior=row_col_list_to_points(external, flip_row_col_order=True),
                                       interior=[row_col_list_to_points(x, flip_row_col_order=True) for x in internals])
                    res.append(new_poly)

        offset_row, offset_col = origin.row, origin.col
        res = [x.translate(offset_row, offset_col) for x in res]
        return res

    def bitwise_mask(self, full_target_mask: np.ndarray, bit_op) -> Bitmap:
        """
        Make bitwise operations between a given numpy array and Bitmap.

        :param full_target_mask: Input numpy array.
        :type full_target_mask: np.ndarray
        :param bit_op: Type of bitwise operation(and, or, not, xor), uses `numpy logic <https://numpy.org/doc/stable/reference/routines.logic.html>`_ functions.
        :type bit_op: `Numpy logical operation <https://numpy.org/doc/stable/reference/routines.logic.html#logical-operations>`_
        :return: Bitmap object or empty list
        :rtype: :class:`Bitmap<Bitmap>` or :class:`list`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            mask = np.array([[0, 0, 0, 0, 0],
                            [0, 1, 1, 1, 0],
                            [0, 1, 0, 1, 0],
                            [0, 1, 1, 1, 0],
                            [0, 0, 0, 0, 0]], dtype=np.bool_)

            figure = sly.Bitmap(mask)

            array = np.array([[0, 0, 0, 0, 0],
                             [0, 1, 1, 1, 0],
                             [0, 0, 1, 0, 0],
                             [0, 0, 1, 0, 0],
                             [0, 0, 0, 0, 0]], dtype=np.bool_)

            bitwise_figure = figure.bitwise_mask(array, np.logical_and)
            print(bitwise_figure.data)
            # Output:
            # [[ True  True  True]
            #  [False False False]
            #  [False  True False]]
        """
        full_size = full_target_mask.shape[:2]
        origin, mask = self.origin, self.data
        full_size_mask = np.full(full_size, False, bool)
        full_size_mask[origin.row:origin.row + mask.shape[0], origin.col:origin.col + mask.shape[1]] = mask

        new_mask = bit_op(full_target_mask, full_size_mask).astype(bool)
        if new_mask.sum() == 0:
            return []
        new_mask = new_mask[origin.row:origin.row + mask.shape[0], origin.col:origin.col + mask.shape[1]]
        return Bitmap(data=new_mask, origin=origin.clone())

    @classmethod
    def allowed_transforms(cls):
        from supervisely.geometry.any_geometry import AnyGeometry
        from supervisely.geometry.polygon import Polygon
        from supervisely.geometry.rectangle import Rectangle
        return [AnyGeometry, Polygon, Rectangle]



# coding: utf-8
import unittest
import numpy as np
import json

from supervisely.geometry.point_location import PointLocation, row_col_list_to_points, points_to_row_col_list
from supervisely.geometry.point import Point
from supervisely.geometry.bitmap import Bitmap
from supervisely.geometry.multichannel_bitmap import MultichannelBitmap
from supervisely.geometry.polygon import Polygon
from supervisely.geometry.rectangle import Rectangle
from supervisely.geometry.image_rotator import ImageRotator
from supervisely.imaging.image import KEEP_ASPECT_RATIO
from supervisely.geometry.constants import POINTS, EXTERIOR, INTERIOR, BITMAP, DATA, ORIGIN


class PointTest(unittest.TestCase):
    def setUp(self):
        self.point = Point(row=10,  col=5)

    def assertPointEquals(self, point, row, col):
        self.assertIsInstance(point, Point)
        self.assertEqual(point.row, row)
        self.assertEqual(point.col, col)

    def test_empty_crop(self):
        crop_rect = Rectangle(100, 100, 200, 200)
        res_geoms = self.point.crop(crop_rect)
        self.assertEqual(len(res_geoms), 0)

    def test_crop(self):
        crop_rect = Rectangle(0, 0, 100, 100)
        res_geoms = self.point.crop(crop_rect)
        self.assertEqual(len(res_geoms), 1)
        res_point = res_geoms[0]
        self.assertPointEquals(res_point, self.point.row, self.point.col)

    def test_relative_crop(self):
        crop_rect = Rectangle(5, 5, 100, 100)
        res_geoms = self.point.relative_crop(crop_rect)
        self.assertEqual(len(res_geoms), 1)
        res_point = res_geoms[0]
        self.assertPointEquals(res_point, 5, 0)

    def test_rotate(self):
        rotator = ImageRotator((21, 21), 90)  # Center pixel is (10, 10)
        rot_point = self.point.rotate(rotator)
        self.assertPointEquals(rot_point, 15, 10)

    def test_resize(self):
        in_size = (20, 40)
        out_size = (30, KEEP_ASPECT_RATIO)
        res_point = self.point.resize(in_size, out_size)
        self.assertPointEquals(res_point, 15, 8)

    def test_scale(self):
        factor = 2.7
        res_point = self.point.scale(factor)
        self.assertPointEquals(res_point, 27, 14)

    def test_translate(self):
        drow = 21
        dcol = -9
        res_point = self.point.translate(drow=drow, dcol=dcol)
        self.assertPointEquals(res_point, 31, -4)

    def test_fliplr(self):
        imsize = (110, 100)
        res_point = self.point.fliplr(imsize)
        self.assertPointEquals(res_point, self.point.row, 95)

    def test_flipud(self):
        imsize = (90, 100)
        res_point = self.point.flipud(imsize)
        self.assertPointEquals(res_point, 80, self.point.col)

    def test_area(self):
        area = self.point.area
        self.assertIsInstance(area, float)
        self.assertEqual(area, 0.0)

    def test_to_bbox(self):
        rect = self.point.to_bbox()
        self.assertPointEquals(self.point, rect.top, rect.left)
        self.assertPointEquals(self.point, rect.bottom, rect.right)

    def test_clone(self):
        res_point = self.point.clone()
        self.assertPointEquals(res_point, self.point.row, self.point.col)
        self.assertIsNot(res_point, self.point)

    def test_from_json(self):
        packed_obj = {
            'some_stuff': 'aaa',
            POINTS: {
                EXTERIOR: [[17, 3]],
                INTERIOR: []
            }
        }
        res_point = Point.from_json(packed_obj)
        self.assertIsInstance(res_point, Point)
        self.assertEqual(res_point.row, 3)
        self.assertEqual(res_point.col, 17)

    def test_to_json(self):
        res_obj = self.point.to_json()
        expected_dict = {
            POINTS: {
                EXTERIOR: [[5, 10]],
                INTERIOR: []
            }
        }
        self.assertDictEqual(res_obj, expected_dict)


class RectangleTest(unittest.TestCase):
    def setUp(self):
        self.rect = Rectangle(top=5, left=10, bottom=30, right=30)

    def assertRectEquals(self, rect, top, left, bottom, right):
        self.assertIsInstance(rect, Rectangle)
        self.assertEqual(rect.top, top)
        self.assertEqual(rect.left, left)
        self.assertEqual(rect.bottom, bottom)
        self.assertEqual(rect.right, right)

    def test_empty_crop(self):
        crop_rect = Rectangle(100, 100, 200, 200)
        res_geoms = self.rect.crop(crop_rect)
        self.assertEqual(len(res_geoms), 0)

    def test_crop(self):
        crop_rect = Rectangle(0, 0, 100, 100)
        res_geoms = self.rect.crop(crop_rect)
        self.assertEqual(len(res_geoms), 1)
        res_rect = res_geoms[0]
        self.assertRectEquals(res_rect, self.rect.top, self.rect.left, self.rect.bottom, self.rect.right)

    def test_relative_crop(self):
        crop_rect = Rectangle(3, 4, 100, 100)
        res_geoms = self.rect.relative_crop(crop_rect)
        self.assertEqual(len(res_geoms), 1)
        res_rect = res_geoms[0]
        self.assertRectEquals(res_rect, 2, 6, 27, 26)

    def test_rotate(self):
        imsize = (101, 101)
        rotator = ImageRotator(imsize=imsize, angle_degrees_ccw=90)
        res_rect = self.rect.rotate(rotator)
        self.assertRectEquals(res_rect, 70, 5, 90, 30)

    def test_resize(self):
        in_size = (100, 100)
        out_size = (200, 150)
        res_rect = self.rect.resize(in_size, out_size)
        self.assertRectEquals(res_rect, 10, 15, 60, 45)

    def test_scale(self):
        factor = 1.3
        res_rect = self.rect.scale(factor)
        self.assertRectEquals(res_rect, 6, 13, 39, 39)

    def test_translate(self):
        drows = 8
        dcols = 259
        res_rect = self.rect.translate(drows, dcols)
        self.assertRectEquals(res_rect, 13, 269, 38, 289)

    def test_fliplr(self):
        im_size = (100, 200)
        res_rect = self.rect.fliplr(im_size)
        self.assertRectEquals(res_rect, 5, 170, 30, 190)

    def test_flipud(self):
        im_size = (100, 200)
        res_rect = self.rect.flipud(im_size)
        self.assertRectEquals(res_rect, 70, 10, 95, 30)

    def test_area(self):
        area = self.rect.area
        self.assertIsInstance(area, float)
        self.assertEqual(area, 546.0)

    def test_to_bbox(self):
        res_rect = self.rect.to_bbox()
        self.assertRectEquals(res_rect, self.rect.top, self.rect.left, self.rect.bottom, self.rect.right)

    def test_clone(self):
        res_rect = self.rect.clone()
        self.assertRectEquals(res_rect, self.rect.top, self.rect.left, self.rect.bottom, self.rect.right)
        self.assertIsNot(res_rect, self.rect)

    def test_from_json(self):
        packed_obj = {
            'some_stuff': 'aaa',
            POINTS: {
                EXTERIOR: [[17, 3], [34, 45]],
                INTERIOR: []
            }
        }
        res_rect = Rectangle.from_json(packed_obj)
        self.assertRectEquals(res_rect, 3, 17, 45, 34)

    def test_to_json(self):
        res_obj = self.rect.to_json()
        expected_dict = {
            POINTS: {
                EXTERIOR: [[10, 5], [30, 30]],
                INTERIOR: []
            }
        }
        self.assertDictEqual(res_obj, expected_dict)

    def test_draw(self):
        rect = Rectangle(1, 1, 3, 3)

        bitmap_1 = np.zeros((5, 5), dtype=np.uint8)
        rect.draw_contour(bitmap_1, 1)
        expected_mask_1 = np.array([[0, 0, 0, 0, 0],
                                    [0, 1, 1, 1, 0],
                                    [0, 1, 0, 1, 0],
                                    [0, 1, 1, 1, 0],
                                    [0, 0, 0, 0, 0]], dtype=np.uint8)
        self.assertTrue(np.array_equal(bitmap_1, expected_mask_1))

        bitmap_2 = np.zeros((5, 5), dtype=np.uint8)
        rect.draw(bitmap_2, 1)
        expected_mask_2 = np.array([[0, 0, 0, 0, 0],
                                    [0, 1, 1, 1, 0],
                                    [0, 1, 1, 1, 0],
                                    [0, 1, 1, 1, 0],
                                    [0, 0, 0, 0, 0]], dtype=np.uint8)
        self.assertTrue(np.array_equal(bitmap_2, expected_mask_2))


class PolygonTest(unittest.TestCase):
    def setUp(self):
        self.exterior = [[10, 10], [40, 10], [30, 40], [10, 30]]
        self.interiors = [[[20, 20], [30, 20], [30, 30], [20, 30]]]
        self.poly = Polygon(exterior=row_col_list_to_points(self.exterior, flip_row_col_order=True),
                            interior=[row_col_list_to_points(self.interiors[0], flip_row_col_order=True)])

    def assertPolyEquals(self, poly, exterior, interiors):
        self.assertIsInstance(poly, Polygon)
        self.assertCountEqual(points_to_row_col_list(poly.exterior, flip_row_col_order=True), exterior)
        self.assertEqual(len(poly.interior), len(interiors))
        for p_interior, interior in zip(poly.interior, interiors):
            self.assertCountEqual(points_to_row_col_list(p_interior, flip_row_col_order=True), interior)

    def test_empty_crop(self):
        crop_rect = Rectangle(100, 100, 200, 200)
        res_geoms = self.poly.crop(crop_rect)
        self.assertEqual(len(res_geoms), 0)

    def test_crop(self):
        crop_rect = Rectangle(25, 0, 200, 200)
        res_geoms = self.poly.crop(crop_rect)
        self.assertEqual(len(res_geoms), 1)
        crop = res_geoms[0]
        self.assertPolyEquals(crop,
                              [[10, 25], [20, 25], [20, 30], [30, 30], [30, 25], [35, 25], [30, 40], [10, 30]],
                              [])

    def test_complex_crop(self):
        # Crop generate GeometryCollection here
        exterior = [[0, 0], [0, 3], [5, 8], [5, 9], [5, 10], [0, 15], [10, 20], [0, 25], [20, 25], [20, 0]]
        interiors = [[[2, 2], [4, 4], [4, 2]]]

        poly = Polygon(exterior=row_col_list_to_points(exterior, flip_row_col_order=True),
                       interior=[row_col_list_to_points(interior, flip_row_col_order=True) for interior in interiors])

        crop_rect = Rectangle(0, 0, 30, 5)
        res_geoms = poly.crop(crop_rect)
        self.assertEqual(len(res_geoms), 3)
        self.assertPolyEquals(res_geoms[0],
                              [[0, 0], [5, 0], [5, 8], [0, 3]], interiors)

    def test_crop_by_border(self):
        exterior = [[10, 10], [40, 10], [40, 40], [10, 40]]
        interiors = [[[11, 11], [11, 20], [20, 11]], [[20, 20], [21, 20], [20, 21]]]
        poly = Polygon(exterior=row_col_list_to_points(exterior, flip_row_col_order=True),
                       interior=[row_col_list_to_points(interior, flip_row_col_order=True) for interior in interiors])

        crop_rect = Rectangle(0, 0, 100, 10)
        res_geoms = poly.crop(crop_rect)
        self.assertEqual(len(res_geoms), 0)

    def test_relative_crop(self):
        crop_rect = Rectangle(25, 0, 200, 200)
        res_geoms = self.poly.relative_crop(crop_rect)
        self.assertEqual(len(res_geoms), 1)
        crop = res_geoms[0]
        self.assertPolyEquals(crop, [[10, 0], [20, 0], [20, 5], [30, 5], [30, 0], [35, 0], [30, 15], [10, 5]], [])

    def test_rotate(self):
        imsize = (101, 101)
        rotator = ImageRotator(imsize=imsize, angle_degrees_ccw=-90)
        res_poly = self.poly.rotate(rotator)
        self.assertPolyEquals(res_poly,
                              [[90, 10], [90, 40], [60, 30], [70, 10]],
                              [[[80, 20], [80, 30], [70, 30], [70, 20]]])

    def test_resize(self):
        in_size = (100, 100)
        out_size = (200, 150)
        res_poly = self.poly.resize(in_size, out_size)
        self.assertPolyEquals(res_poly,
                              [[15, 20], [60, 20], [45, 80], [15, 60]],
                              [[[30, 40], [45, 40], [45, 60], [30, 60]]])

    def test_translate(self):
        drow = 5
        dcol = 10
        res_poly = self.poly.translate(drow, dcol)
        self.assertPolyEquals(res_poly,
                              [[20, 15], [50, 15], [40, 45], [20, 35]],
                              [[[30, 25], [40, 25], [40, 35], [30, 35]]])

    def test_fliplr(self):
        imsize = (100, 200)
        res_poly = self.poly.fliplr(imsize)
        self.assertPolyEquals(res_poly,
                              [[190, 10], [160, 10], [170, 40], [190, 30]],
                              [[[180, 20], [170, 20], [170, 30], [180, 30]]])

    def test_area(self):  # @TODO: only exterior area
        area = self.poly.area
        self.assertIsInstance(area, float)
        self.assertEqual(area, 650.0)

    def test_to_bbox(self):
        res_rect = self.poly.to_bbox()
        self.assertIsInstance(res_rect, Rectangle)
        self.assertEqual(res_rect.top, 10)
        self.assertEqual(res_rect.left, 10)
        self.assertEqual(res_rect.right, 40)
        self.assertEqual(res_rect.bottom, 40)

    def test_from_json(self):
        packed_obj = {
            'some_stuff': 'aaa',
            POINTS: {
                EXTERIOR: self.exterior,
                INTERIOR: self.interiors
            }
        }
        res_poly = Polygon.from_json(packed_obj)
        self.assertPolyEquals(res_poly, self.exterior, self.interiors)

    def test_to_json(self):
        res_obj = self.poly.to_json()
        expected_dict = {
            POINTS: {
                EXTERIOR: self.exterior,
                INTERIOR: self.interiors
            }
        }
        self.assertDictEqual(res_obj, expected_dict)

    def test_clone(self):
        res_poly = self.poly.clone()
        self.assertIsInstance(res_poly, Polygon)
        self.assertTrue(np.array_equal(res_poly.exterior_np, self.poly.exterior_np))
        self.assertTrue(np.array_equal(res_poly.interior_np, self.poly.interior_np))
        self.assertIsNot(res_poly, self.poly)

    def test_draw(self):
        def draw_mask(exterior_p, interior_p, h, w):
            mask = np.zeros((h, w), dtype=np.uint8)
            poly = Polygon(exterior=row_col_list_to_points(exterior_p),
                           interior=[row_col_list_to_points(interior) for interior in interior_p])
            poly.draw(mask, color=1)
            return mask

        # Test 1 - draw interior triangles
        exterior = [[0, 0], [0, 6], [4, 6], [4, 0]]
        interiors = [[[1, 1], [1, 3], [3, 1]],  # clockwise
                     [[1, 5], [3, 3], [3, 5]]]  # counterclockwise
        mask_1 = draw_mask(exterior, interiors, h=5, w=7)
        expected_mask = np.array([[1, 1, 1, 1, 1, 1, 1],
                                  [1, 0, 0, 0, 1, 0, 1],
                                  [1, 0, 0, 1, 0, 0, 1],
                                  [1, 0, 1, 0, 0, 0, 1],
                                  [1, 1, 1, 1, 1, 1, 1]], dtype=np.uint8)

        self.assertTrue(np.array_equal(mask_1, expected_mask))

        # Test 1 - draw interior sandglass (bad poly case)
        exterior = [[0, 0], [0, 7], [7, 7], [7, 0]]
        interiors = [[[1, 1], [5, 5], [1, 5], [5, 1]]]  # sandglass
        mask_2 = draw_mask(exterior, interiors, h=7, w=7)
        expected_mask = np.array([[1, 1, 1, 1, 1, 1, 1],
                                  [1, 0, 1, 1, 1, 0, 1],
                                  [1, 0, 0, 1, 0, 0, 1],
                                  [1, 0, 0, 0, 0, 0, 1],
                                  [1, 0, 0, 1, 0, 0, 1],
                                  [1, 0, 1, 1, 1, 0, 1],
                                  [1, 1, 1, 1, 1, 1, 1]], dtype=np.uint8)
        self.assertTrue(np.array_equal(mask_2, expected_mask))

    def test_draw_contour(self):
        exterior = [[0, 0], [0, 6], [6, 6], [6, 0]]
        interiors = [[[2, 2], [2, 4], [4, 4], [4, 2]]]
        poly = Polygon(exterior=row_col_list_to_points(exterior),
                       interior=[row_col_list_to_points(interior) for interior in interiors])

        bitmap_1 = np.zeros((7, 7), dtype=np.uint8)
        poly.draw_contour(bitmap_1, color=1)

        expected_mask = np.array([[1, 1, 1, 1, 1, 1, 1],
                                  [1, 0, 0, 0, 0, 0, 1],
                                  [1, 0, 1, 1, 1, 0, 1],
                                  [1, 0, 1, 0, 1, 0, 1],
                                  [1, 0, 1, 1, 1, 0, 1],
                                  [1, 0, 0, 0, 0, 0, 1],
                                  [1, 1, 1, 1, 1, 1, 1]], dtype=np.uint8)
        self.assertTrue(np.array_equal(bitmap_1, expected_mask))

        # Extended test
        exterior = [[0, 0], [0, 6], [10, 6], [10, 0]]
        interiors = [[[1, 1], [1, 2], [2, 1]],
                     [[2, 4], [3, 5], [4, 4], [3, 3]],
                     [[6, 2], [8, 2], [8, 4], [6, 4]]]
        poly = Polygon(exterior=row_col_list_to_points(exterior),
                       interior=[row_col_list_to_points(interior) for interior in interiors])

        bitmap_2 = np.zeros((11, 7), dtype=np.uint8)
        poly.draw_contour(bitmap_2, color=1)

        expected_mask = np.array([[1, 1, 1, 1, 1, 1, 1],
                                  [1, 1, 1, 0, 0, 0, 1],
                                  [1, 1, 0, 0, 1, 0, 1],
                                  [1, 0, 0, 1, 0, 1, 1],
                                  [1, 0, 0, 0, 1, 0, 1],
                                  [1, 0, 0, 0, 0, 0, 1],
                                  [1, 0, 1, 1, 1, 0, 1],
                                  [1, 0, 1, 0, 1, 0, 1],
                                  [1, 0, 1, 1, 1, 0, 1],
                                  [1, 0, 0, 0, 0, 0, 1],
                                  [1, 1, 1, 1, 1, 1, 1]], dtype=np.uint8)

        self.assertTrue(np.array_equal(bitmap_2, expected_mask))


class BitmapTest(unittest.TestCase):
    def setUp(self):
        self.origin = PointLocation(0, 4)
        self.mask = np.array([[0, 0, 0, 1, 0, 0, 0],
                              [0, 0, 1, 1, 1, 0, 0],
                              [0, 1, 0, 1, 0, 1, 0],
                              [0, 0, 0, 1, 0, 0, 0],
                              [0, 0, 0, 1, 0, 0, 0],
                              [0, 0, 0, 1, 0, 0, 0],
                              [0, 0, 0, 1, 0, 0, 0]], dtype=np.bool)
        self.mask_no_margins = self.mask[:, 1:-1]
        self.bitmap = Bitmap(data=self.mask, origin=self.origin)

    def assertBitmapEquals(self, bitmap, origin_row, origin_col, mask):
        self.assertIsInstance(bitmap, Bitmap),
        self.assertEqual(bitmap.origin.row, origin_row)
        self.assertEqual(bitmap.origin.col, origin_col)
        self.assertListEqual(bitmap.data.tolist(), mask.tolist())

    def test_rotate(self):
        in_size = (15, 15)
        rotator = ImageRotator(imsize=in_size, angle_degrees_ccw=90)
        res_bitmap = self.bitmap.rotate(rotator)
        expected_mask = np.array([[0, 0, 1, 0, 0, 0, 0],
                                  [0, 1, 0, 0, 0, 0, 0],
                                  [1, 1, 1, 1, 1, 1, 1],
                                  [0, 1, 0, 0, 0, 0, 0],
                                  [0, 0, 1, 0, 0, 0, 0]], dtype=np.bool)
        self.assertListEqual(res_bitmap.data.tolist(), expected_mask.tolist())

    def test_empty_crop(self):
        crop_rect = Rectangle(0, 0, 4, 4)
        res_geoms = self.bitmap.crop(crop_rect)
        self.assertEqual(len(res_geoms), 0)

    def test_crop(self):  # @TODO: mb delete compress while cropping
        crop_rect = Rectangle(0, 0, 8, 8)
        res_geoms = self.bitmap.crop(crop_rect)
        self.assertEqual(len(res_geoms), 1)
        res_bitmap = res_geoms[0]
        res_mask = np.array([[0, 0, 1, 0],
                             [0, 1, 1, 1],
                             [1, 0, 1, 0],
                             [0, 0, 1, 0],
                             [0, 0, 1, 0],
                             [0, 0, 1, 0],
                             [0, 0, 1, 0]], dtype=np.bool)
        self.assertBitmapEquals(res_bitmap, 0, 5, res_mask)

    def test_translate(self):
        drows = 10
        dcols = 7
        res_bitmap = self.bitmap.translate(drows, dcols)
        self.assertBitmapEquals(res_bitmap, 10, 12, self.mask_no_margins)

    def test_area(self):
        area = self.bitmap.area
        self.assertIsInstance(area, float)
        self.assertEqual(area, 11)

    def test_to_bbox(self):
        res_rect = self.bitmap.to_bbox()
        self.assertIsInstance(res_rect, Rectangle)
        self.assertEqual(res_rect.top, 0)
        self.assertEqual(res_rect.left, 5)
        self.assertEqual(res_rect.right, 9)
        self.assertEqual(res_rect.bottom, 6)

    def test_to_contours(self):
        bitmap = Bitmap(data=np.array([[1, 1, 0, 1, 1, 1],
                                       [1, 1, 0, 1, 0, 1],
                                       [0, 0, 0, 1, 1, 1],
                                       [1, 0, 0, 1, 0, 1],
                                       [1, 0, 0, 1, 1, 1],
                                       [1, 0, 0, 0, 0, 0],
                                       [1, 0, 0, 1, 1, 1]], dtype=np.bool),
                        origin=PointLocation(10, 110))
        polygons = bitmap.to_contours()

        exteriors_points = [np.array([[10, 113], [14, 113], [14, 115], [10, 115]]),
                            np.array([[10, 110], [11, 110], [11, 111], [10, 111]])]

        interiors_points = [[],
                            [np.array([[13, 113], [12, 114], [13, 115], [14, 114]]),
                             np.array([[11, 113], [10, 114], [11, 115], [12, 114]])],
                           []]

        self.assertEqual(len(polygons), 2)
        for polygon, target_exterior, target_interiors in zip(polygons, exteriors_points, interiors_points):
            self.assertTrue(np.equal(polygon.exterior_np, target_exterior).all())
            self.assertTrue(all(np.equal(p_inter, t_inter)
                                for p_inter, t_inter in zip(polygon.interior_np, target_interiors)))
            json.dumps(polygon.to_json())
            self.assertIsInstance(polygon, Polygon)

    def test_from_json(self):
        packed_obj = {
            BITMAP: {
                DATA: 'eJzrDPBz5+WS4mJgYOD19HAJAtLsIMzIDCT/zTk6AUixBfiEuALp////L705/y6QxVgS5BfM4PDsRhqQI+j'
                      'p4hhSMSdZIGFDAkeiQIMDA7sVw125xatvACUZPF39XNY5JTQBADRqHJQ=',
                ORIGIN: [0, 1]
            }
        }
        bitmap = Bitmap.from_json(packed_obj)
        self.assertBitmapEquals(bitmap, 1, 1, self.mask_no_margins)

    def test_resize(self):
        in_size = (20, 20)
        out_size = (40, 40)
        res_bitmap = self.bitmap.resize(in_size, out_size)
        self.assertIsInstance(res_bitmap, Bitmap)

    def test_flipud(self):
        im_size = (20, 20)
        res_bitmap = self.bitmap.flipud(im_size)
        expected_mask = np.array([[0, 0, 1, 0, 0],
                                  [0, 0, 1, 0, 0],
                                  [0, 0, 1, 0, 0],
                                  [0, 0, 1, 0, 0],
                                  [1, 0, 1, 0, 1],
                                  [0, 1, 1, 1, 0],
                                  [0, 0, 1, 0, 0]], dtype=np.bool)
        self.assertBitmapEquals(res_bitmap, 13, 5, expected_mask)

    def test_to_json(self):
        obj = self.bitmap.to_json()
        expected_dict = {
            BITMAP: {
                DATA: 'eJzrDPBz5+WS4mJgYOD19HAJAtKsQMzOyAwkf2WKrgVSbAE+Ia5A+v///0tvzr8LZDGWBPkFMzg8u5EG5Ah'
                      '6ujiGVMxJDkjQSIg4uIChkYEvjXHncvfAUqAkg6ern8s6p4QmAAVvHAE=',
                ORIGIN: [5, 0]
            }
        }
        self.assertDictEqual(obj, expected_dict)


class MultichannelBitmapTest(unittest.TestCase):
    def setUp(self):
        self.origin = PointLocation(row=0, col=4)
        self.data = np.array([[[0.0, 0.1], [0.2, 0.3], [0.4, 0.5]],
                              [[0.6, 0.7], [0.8, 0.9], [1.0, 1.1]]], dtype=np.float64)
        self.bitmap = MultichannelBitmap(data=self.data, origin=self.origin)

    def assertMultichannelBitmapEquals(self, bitmap, origin_row, origin_col, data):
        self.assertIsInstance(bitmap, MultichannelBitmap),
        self.assertEqual(bitmap.origin.row, origin_row)
        self.assertEqual(bitmap.origin.col, origin_col)
        self.assertListEqual(bitmap.data.tolist(), data.tolist())

    def test_rotate(self):
        in_size = (15, 15)
        rotator = ImageRotator(imsize=in_size, angle_degrees_ccw=90)
        res_bitmap = self.bitmap.rotate(rotator)
        expected_data = np.array([[[0.4, 0.5], [1.0, 1.1]],
                                  [[0.2, 0.3], [0.8, 0.9]],
                                  [[0.0, 0.1], [0.6, 0.7]]], dtype=np.float64)
        # TODO origin
        self.assertListEqual(res_bitmap.data.tolist(), expected_data.tolist())

    def test_empty_crop(self):
        crop_rect = Rectangle(top=0, left=0, bottom=10, right=3)
        res_geoms = self.bitmap.crop(crop_rect)
        self.assertEqual(len(res_geoms), 0)

    def test_crop(self):
        crop_rect = Rectangle(top=1, left=0, bottom=10, right=4)
        res_geoms = self.bitmap.crop(crop_rect)
        self.assertEqual(len(res_geoms), 1)
        res_bitmap = res_geoms[0]
        res_data = np.array([[[0.6, 0.7]]], dtype=np.float64)
        self.assertMultichannelBitmapEquals(res_bitmap, 1, 4, res_data)

    def test_translate(self):
        drows = 10
        dcols = 7
        res_bitmap = self.bitmap.translate(drows, dcols)
        self.assertMultichannelBitmapEquals(res_bitmap, 10, 11, self.data)

    def test_area(self):
        area = self.bitmap.area
        self.assertEqual(area, 6)

    def test_to_bbox(self):
        res_rect = self.bitmap.to_bbox()
        self.assertIsInstance(res_rect, Rectangle)
        self.assertEqual(res_rect.top, 0)
        self.assertEqual(res_rect.left, 4)
        self.assertEqual(res_rect.right, 6)
        self.assertEqual(res_rect.bottom, 1)

    def test_resize(self):
        in_size = (20, 20)
        out_size = (40, 40)
        res_bitmap = self.bitmap.resize(in_size, out_size)
        expected_data = np.array([[[0.0, 0.1], [0.0, 0.1], [0.2, 0.3], [0.2, 0.3], [0.4, 0.5], [0.4, 0.5]],
                                  [[0.0, 0.1], [0.0, 0.1], [0.2, 0.3], [0.2, 0.3], [0.4, 0.5], [0.4, 0.5]],
                                  [[0.6, 0.7], [0.6, 0.7], [0.8, 0.9], [0.8, 0.9], [1.0, 1.1], [1.0, 1.1]],
                                  [[0.6, 0.7], [0.6, 0.7], [0.8, 0.9], [0.8, 0.9], [1.0, 1.1], [1.0, 1.1]]],
                                 dtype=np.float64)
        self.assertMultichannelBitmapEquals(res_bitmap, 0, 8, expected_data)

    def test_flipud(self):
        im_size = (20, 20)
        res_bitmap = self.bitmap.flipud(im_size)
        expected_data = np.array([[[0.6, 0.7], [0.8, 0.9], [1.0, 1.1]],
                                  [[0.0, 0.1], [0.2, 0.3], [0.4, 0.5]]], dtype=np.float64)
        self.assertMultichannelBitmapEquals(res_bitmap, 18, 4, expected_data)

    def test_fliplr(self):
        im_size = (20, 20)
        res_bitmap = self.bitmap.fliplr(im_size)
        expected_data = np.array([[[0.4, 0.5], [0.2, 0.3], [0.0, 0.1]],
                                  [[1.0, 1.1], [0.8, 0.9], [0.6, 0.7]]], dtype=np.float64)
        self.assertMultichannelBitmapEquals(res_bitmap, 0, 13, expected_data)


if __name__ == '__main__':
    unittest.main()



# coding: utf-8

from supervisely.video_annotation.frame import Frame
from supervisely.volume_annotation.volume_figure import VolumeFigure
from supervisely.volume_annotation.constants import FIGURES, INDEX


class Slice(Frame):
    figure_type = VolumeFigure

    @classmethod
    def from_json(cls, data, objects, slices_count=None, key_id_map=None):
        raise NotImplementedError()
        _frame = super().from_json(data, objects, slices_count, key_id_map)
        return cls(index=_frame.index, figures=_frame.figures)

    @classmethod
    def from_json(cls, data, objects, plane_name, slices_count=None, key_id_map=None):
        index = data[INDEX]
        if index < 0:
            raise ValueError("Frame Index have to be >= 0")

        if slices_count is not None:
            if index > slices_count:
                raise ValueError(
                    "Item contains {} frames. Frame index is {}".format(
                        slices_count, index
                    )
                )

        figures = []
        for figure_json in data.get(FIGURES, []):
            figure = cls.figure_type.from_json(
                figure_json, objects, plane_name, index, key_id_map
            )
            figures.append(figure)
        return cls(index=index, figures=figures)



# coding: utf-8
"""collection with :class:`ObjClass<supervisely.annotation.obj_class.ObjClass>` instances"""

# docs
from __future__ import annotations
from typing import List, Optional, Dict
from supervisely import logger
from supervisely.annotation.renamer import Renamer

from collections import defaultdict
from supervisely.collection.key_indexed_collection import KeyIndexedCollection
from supervisely.io.json import JsonSerializable
from supervisely.annotation.obj_class import ObjClass
from supervisely.imaging.color import rgb2hex, hex2rgb


class ObjClassCollection(KeyIndexedCollection, JsonSerializable):
    """
    Collection with :class:`ObjClass<supervisely.annotation.obj_class.ObjClass>` instances. :class:`ObjClassCollection<ObjClassCollection>` object is immutable.

    :raises: :class:`DuplicateKeyError` if instance with given name already exist
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        # Create ObjClass (see class ObjClass for more information)
        class_lemon = sly.ObjClass('lemon', sly.Rectangle)
        class_kiwi = sly.ObjClass('kiwi', sly.Bitmap)

        class_arr = [class_lemon, class_kiwi]

        # Create ObjClassCollection from ObjClasses
        classes = sly.ObjClassCollection(class_arr)

        # Add items to ObjClassCollection
        class_potato = sly.ObjClass('potato', sly.Bitmap)

        # Remember that ObjClassCollection is immutable, and we need to assign new instance of ObjClassCollection to a new variable
        classes = classes.add(class_potato)

        # You can also add multiple items to collection
        class_cabbage = sly.ObjClass('cabbage', sly.Rectangle)
        class_carrot = sly.ObjClass('carrot', sly.Bitmap)
        class_turnip = sly.ObjClass('turnip', sly.Polygon)

        classes = classes.add_items([class_cabbage, class_carrot, class_turnip])

        # Has key, checks if given key exist in collection
        classes.has_key("cabbage")
        # Output: True

        # Intersection, finds intersection of given list of instances with collection items
        class_dog = sly.ObjClass('dog', sly.Rectangle)
        class_cat = sly.ObjClass('cat', sly.Rectangle)
        class_turtle = sly.ObjClass('turtle', sly.Rectangle)

        classes_animals = sly.ObjClassCollection([class_dog, class_cat, class_turtle])

        classes_intersections = classes.intersection(classes_animals)
        print(classes_intersections.to_json())
        # Output: []

        # Let's add the potato ObjClass from another collection and compare them again
        classes_animals = classes_animals.add(class_potato)

        classes_intersections = classes.intersection(classes_animals)
        print(classes_intersections.to_json())
        # Output: [
        #     {
        #         "title":"potato",
        #         "shape":"bitmap",
        #         "color":"#8A570F",
        #         "geometry_config":{},
        #         "hotkey":""
        #     }
        # ]

        # Difference, finds difference between collection and given list of ObjClass or ObjClassCollection
        class_car = sly.ObjClass('car', sly.Rectangle)
        class_bicycle = sly.ObjClass('bicycle', sly.Rectangle)

        classes_vehicles = sly.ObjClassCollection([class_car, class_bicycle])

        class_pedestrian = sly.ObjClass('pedestrian', sly.Rectangle)
        class_road = sly.ObjClass('road', sly.Rectangle)

        difference = classes_vehicles.difference([class_pedestrian, class_road])
        print(difference.to_json())
        # Output: [
        #     {
        #         "title":"car",
        #         "shape":"rectangle",
        #         "color":"#8A0F3B",
        #         "geometry_config":{},
        #         "hotkey":""
        #     },
        #     {
        #         "title":"bicycle",
        #         "shape":"rectangle",
        #         "color":"#0F8A1F",
        #         "geometry_config":{},
        #         "hotkey":""
        #     }
        # ]

        # Merge, merges collection and given list of ObjClasses
        c_1 = sly.ObjClassCollection([class_car, class_bicycle])
        c_2 = sly.ObjClassCollection([class_pedestrian, class_road])

        с_3 = c_1.merge(c_2)
        print(с_3.to_json())
        # Output: [
        #     {
        #         "title":"pedestrian",
        #         "shape":"rectangle",
        #         "color":"#8A0F27",
        #         "geometry_config":{},
        #         "hotkey":""
        #     },
        #     {
        #         "title":"road",
        #         "shape":"rectangle",
        #         "color":"#8A620F",
        #         "geometry_config":{},
        #         "hotkey":""
        #     },
        #     {
        #         "title":"car",
        #         "shape":"rectangle",
        #         "color":"#8A0F3B",
        #         "geometry_config":{},
        #         "hotkey":""
        #     },
        #     {
        #         "title":"bicycle",
        #         "shape":"rectangle",
        #         "color":"#0F8A1F",
        #         "geometry_config":{},
        #         "hotkey":""
        #     }
        # ]

        # Merge will raise ValueError if item name from given list is in collection but items in both are different
        class_bicycle_1 = sly.ObjClass('bicycle', sly.Rectangle)
        class_bicycle_2 = sly.ObjClass('bicycle', sly.Bitmap)

        classes_1 = sly.ObjClassCollection([class_bicycle_1])
        classes_2 = sly.ObjClassCollection([class_bicycle_2])

        test_merge = classes_1.merge(classes_2)
        # Output: ValueError: Error during merge for key 'bicycle': values are different

        # Let's try to create now a collection where ObjClasses have identical names
        class_cow = sly.ObjClass('cow', sly.Rectangle)
        class_chicken = sly.ObjClass('cow', sly.Rectangle)

        test_classes = sly.ObjClassCollection([class_cow, class_chicken])
        # Output: DuplicateKeyError: "Key 'cow' already exists"
    """

    item_type = ObjClass

    def to_json(self) -> List[Dict]:
        """
        Convert the ObjClassCollection to a list of json dicts. Read more about `Supervisely format <https://docs.supervise.ly/data-organization/00_ann_format_navi>`_.

        :return: List of dicts in json format
        :rtype: :class:`List[dict]`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            class_lemon = sly.ObjClass('lemon', sly.Rectangle)
            class_kiwi = sly.ObjClass('kiwi', sly.Bitmap)

            # Add ObjClasses to ObjClassCollection
            classes = sly.ObjClassCollection([class_lemon, class_kiwi])

            classes_json = classes.to_json()
            print(classes_json)
            # Output: [
            #      {
            #           "title": "lemon",
            #           "shape": "rectangle",
            #           "color": "#300F8A",
            #           "geometry_config": {},
            #           "hotkey": ""
            #      },
            #      {
            #           "title": "kiwi",
            #           "shape": "bitmap",
            #           "color": "#7C0F8A",
            #           "geometry_config": {},
            #           "hotkey": ""
            #      }
            # ]
        """
        return [obj_class.to_json() for obj_class in self]

    @classmethod
    def from_json(cls, data: List[Dict]) -> ObjClassCollection:
        """
        Convert a list with dicts in json format to ObjClassCollection. Read more about `Supervisely format <https://docs.supervise.ly/data-organization/00_ann_format_navi>`_.

        :param data: List with dicts in json format.
        :type data: List[dict]
        :return: ObjClassCollection object
        :rtype: :class:`ObjClassCollection<ObjClassCollection>`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            data = [
                 {
                      "title": "lemon",
                      "shape": "rectangle",
                      "color": "#300F8A",
                      "hotkey": ""
                 },
                 {
                      "title": "kiwi",
                      "shape": "bitmap",
                      "color": "#7C0F8A",
                      "hotkey": ""
                 }
            ]

            classes = sly.ObjClassCollection.from_json(data)
        """
        obj_classes = [ObjClass.from_json(obj_class_json) for obj_class_json in data]
        return cls(obj_classes)

    def validate_classes_colors(self, logger: Optional[logger] = None) -> str or None:
        """
        Checks for unique colors in the ObjClassCollection.

        :param logger: Input logger.
        :type logger: logger, optional
        :return: Notification if there are objects with the same colors, otherwise :class:`None`
        :rtype: :class:`str` or :class:`NoneType`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            # Let's create 2 ObjClasses with the same color
            class_lemon = sly.ObjClass('lemon', sly.Rectangle, [0, 0, 0])
            class_kiwi = sly.ObjClass('kiwi', sly.Bitmap, [0, 0, 0])

            # Add them to ObjClassCollection
            classes = sly.ObjClassCollection([class_lemon, class_kiwi])

            print(classes.validate_classes_colors())
            # Output: Classes ['lemon', 'kiwi'] have the same RGB color = [0, 0, 0]

            # Now let's change colors of our ObjClasses
            class_lemon = sly.ObjClass('lemon', sly.Rectangle, [255, 0, 0])
            class_kiwi = sly.ObjClass('kiwi', sly.Bitmap, [0, 0, 255])

            classes = sly.ObjClassCollection([class_lemon, class_kiwi])

            print(classes.validate_classes_colors())
            # Output: None
        """
        color_names = defaultdict(list)
        for obj_class in self:
            hex = rgb2hex(obj_class.color)
            color_names[hex].append(obj_class.name)

        class_colors_notify = None
        for hex_color, class_names in color_names.items():
            if len(class_names) > 1:
                warn_str = "Classes {!r} have the same RGB color = {!r}".format(class_names, hex2rgb(hex_color))
                if logger is not None:
                    logger.warn(warn_str)
                if class_colors_notify is None:
                    class_colors_notify = ""
                class_colors_notify += warn_str + '\n\n'
        return class_colors_notify


def make_renamed_classes(src_obj_classes: ObjClassCollection, renamer: Renamer, skip_missing: Optional[bool]=False) -> ObjClassCollection:
    renamed_classes = []
    for src_cls in src_obj_classes:
        renamed_name = renamer.rename(src_cls.name)
        if renamed_name is not None:
            renamed_classes.append(src_cls.clone(name=renamed_name))
        elif not skip_missing:
            raise KeyError('Object class name {} could not be mapped to a destination name.'.format(src_cls.name))
    return ObjClassCollection(items=renamed_classes)
